from unittest import TestCase
from niaaml import Pipeline, OptimizationStats
from niaaml.classifiers import RandomForest, AdaBoost
from niaaml.preprocessing.feature_selection import SelectKBest, SelectPercentile
from niaaml.preprocessing.feature_transform import StandardScaler, Normalizer
from niaaml.data import CSVDataReader
import os
import numpy
import tempfile
import pandas


class PipelineTestCase(TestCase):
    def test_pipeline_optimize_works_fine(self):
        pipeline = Pipeline(
            feature_selection_algorithm=SelectKBest(),
            feature_transform_algorithm=Normalizer(),
            classifier=RandomForest(),
        )

        data_reader = CSVDataReader(
            src=os.path.dirname(os.path.abspath(__file__))
            + "/tests_files/dataset_header_classes.csv",
            has_header=True,
            contains_classes=True,
        )

        self.assertIsInstance(pipeline.get_classifier(), RandomForest)
        self.assertIsInstance(pipeline.get_feature_selection_algorithm(), SelectKBest)
        self.assertIsInstance(pipeline.get_feature_transform_algorithm(), Normalizer)

        accuracy = pipeline.optimize(
            data_reader.get_x(),
            data_reader.get_y(),
            20,
            40,
            "ParticleSwarmAlgorithm",
            "Accuracy",
        )

        if accuracy != float("inf"):
            self.assertGreaterEqual(accuracy, -1.0)
            self.assertLessEqual(accuracy, 0.0)

        self.assertIsInstance(pipeline.get_classifier(), RandomForest)
        self.assertIsInstance(pipeline.get_feature_selection_algorithm(), SelectKBest)
        self.assertIsInstance(pipeline.get_feature_transform_algorithm(), Normalizer)

    def test_pipeline_run_works_fine(self):
        pipeline = Pipeline(
            feature_selection_algorithm=SelectKBest(),
            feature_transform_algorithm=Normalizer(),
            classifier=RandomForest(),
        )

        data_reader = CSVDataReader(
            src=os.path.dirname(os.path.abspath(__file__))
            + "/tests_files/dataset_header_classes.csv",
            has_header=True,
            contains_classes=True,
        )
        pipeline.optimize(
            data_reader.get_x(),
            data_reader.get_y(),
            20,
            40,
            "ParticleSwarmAlgorithm",
            "Accuracy",
        )
        predicted = pipeline.run(
            pandas.DataFrame(
                numpy.random.uniform(
                    low=0.0, high=15.0, size=(30, data_reader.get_x().shape[1])
                )
            )
        )

        self.assertEqual(predicted.shape, (30,))

        s1 = set(data_reader.get_y())
        s2 = set(predicted)
        self.assertTrue(s2.issubset(s1))
        self.assertTrue(len(s2) > 0 and len(s2) <= 2)

    def test_pipeline_export_works_fine(self):
        pipeline = Pipeline(
            feature_selection_algorithm=SelectKBest(),
            feature_transform_algorithm=Normalizer(),
            classifier=RandomForest(),
        )

        with tempfile.TemporaryDirectory() as tmp:
            pipeline.export(os.path.join(tmp, "pipeline"))
            self.assertTrue(os.path.exists(os.path.join(tmp, "pipeline.ppln")))
            self.assertEqual(1, len([name for name in os.listdir(tmp)]))

            pipeline.export(os.path.join(tmp, "pipeline.ppln"))
            self.assertTrue(os.path.exists(os.path.join(tmp, "pipeline.ppln")))
            self.assertEqual(1, len([name for name in os.listdir(tmp)]))

    def test_pipeline_export_text_works_fine(self):
        pipeline = Pipeline(
            feature_selection_algorithm=SelectKBest(),
            feature_transform_algorithm=Normalizer(),
            classifier=RandomForest(),
        )

        with tempfile.TemporaryDirectory() as tmp:
            pipeline.export_text(os.path.join(tmp, "pipeline"))
            self.assertTrue(os.path.exists(os.path.join(tmp, "pipeline.txt")))
            self.assertEqual(1, len([name for name in os.listdir(tmp)]))

            pipeline.export_text(os.path.join(tmp, "pipeline.txt"))
            self.assertTrue(os.path.exists(os.path.join(tmp, "pipeline.txt")))
            self.assertEqual(1, len([name for name in os.listdir(tmp)]))

        self.assertIsNotNone(pipeline.to_string())
        self.assertGreater(len(pipeline.to_string()), 0)

    def test_pipeline_setters_work_fine(self):
        pipeline = Pipeline(
            feature_selection_algorithm=SelectKBest(),
            feature_transform_algorithm=Normalizer(),
            classifier=RandomForest(),
        )

        pipeline.set_classifier(AdaBoost())
        pipeline.set_feature_selection_algorithm(SelectPercentile())
        pipeline.set_feature_transform_algorithm(StandardScaler())
        pipeline.set_selected_features_mask(numpy.ones([1, 1, 0, 0], dtype=bool))

        self.__y = numpy.array(
            [
                "Class 1",
                "Class 1",
                "Class 1",
                "Class 2",
                "Class 1",
                "Class 2",
                "Class 2",
                "Class 2",
                "Class 2",
                "Class 1",
                "Class 1",
                "Class 2",
                "Class 1",
                "Class 2",
                "Class 1",
                "Class 1",
                "Class 1",
                "Class 1",
                "Class 2",
                "Class 1",
            ]
        )
        self.__predicted = numpy.array(
            [
                "Class 1",
                "Class 1",
                "Class 1",
                "Class 2",
                "Class 2",
                "Class 2",
                "Class 1",
                "Class 1",
                "Class 1",
                "Class 2",
                "Class 1",
                "Class 1",
                "Class 2",
                "Class 2",
                "Class 1",
                "Class 2",
                "Class 1",
                "Class 2",
                "Class 2",
                "Class 2",
            ]
        )
        pipeline.set_stats(OptimizationStats(self.__predicted, self.__y))

        self.assertIsInstance(pipeline.get_classifier(), AdaBoost)
        self.assertIsInstance(
            pipeline.get_feature_selection_algorithm(), SelectPercentile
        )
        self.assertIsInstance(
            pipeline.get_feature_transform_algorithm(), StandardScaler
        )
        self.assertIsInstance(pipeline.get_stats(), OptimizationStats)



from niapy.algorithms.basic import DifferentialEvolution as DE
from niapy.task import StoppingTask
from niaaml.preprocessing.feature_selection.feature_selection_algorithm import (
    FeatureSelectionAlgorithm,
)
from niaaml.utilities import ParameterDefinition, MinMax
from niaaml.preprocessing.feature_selection._feature_selection_threshold_benchmark import (
    _FeatureSelectionThresholdBenchmark,
)
import numpy

__all__ = ["DifferentialEvolution"]


class DifferentialEvolution(FeatureSelectionAlgorithm):
    r"""Implementation of feature selection using DE algorithm.

    Date:
        2020

    Author:
        Luka Pečnik

    Reference:
        The implementation is adapted according to the following article:
        D. Fister, I. Fister, T. Jagrič, I. Fister Jr., J. Brest. A novel self-adaptive differential evolution for feature selection using threshold mechanism . In: Proceedings of the 2018 IEEE Symposium on Computational Intelligence (SSCI 2018), pp. 17-24, 2018.

    Reference URL:
        http://iztok-jr-fister.eu/static/publications/236.pdf

    License:
        MIT

    See Also:
        * :class:`niaaml.preprocessing.feature_selection.feature_selection_algorithm.FeatureSelectionAlgorithm`
    """
    Name = "Differential Evolution"

    def __init__(self, **kwargs):
        r"""Initialize DE feature selection algorithm."""
        self._params = dict(
            F=ParameterDefinition(MinMax(0.5, 0.9), param_type=float),
            CR=ParameterDefinition(MinMax(0.0, 1.0), param_type=float),
        )
        self.__de = DE(NP=10)

    def set_parameters(self, **kwargs):
        r"""Set the parameters/arguments of the algorithm."""
        kwargs["NP"] = self.__de.NP
        self.__de.set_parameters(**kwargs)

    def __final_output(self, sol):
        r"""Calculate final array of features.

        Arguments:
            sol (numpy.ndarray[float]): Individual of population/ possible solution.

        Returns:
            numpy.ndarray[bool]: Mask of selected features.
        """
        selected = numpy.ones(sol.shape[0] - 1, dtype=bool)
        threshold = sol[sol.shape[0] - 1]
        for i in range(sol.shape[0] - 1):
            if sol[i] < threshold:
                selected[i] = False
        return selected

    def select_features(self, x, y, **kwargs):
        r"""Perform the feature selection process.

        Arguments:
            x (pandas.core.frame.DataFrame): Array of original features.
            y (pandas.core.series.Series) Expected classifier results.

        Returns:
            numpy.ndarray[bool]: Mask of selected features.
        """
        num_features = x.shape[1]
        benchmark = _FeatureSelectionThresholdBenchmark(x, y)
        task = StoppingTask(
            dimension=num_features + 1, max_evals=1000, benchmark=benchmark
        )
        best = self.__de.run(task)
        return self.__final_output(benchmark.get_best_solution())

    def to_string(self):
        r"""User friendly representation of the object.

        Returns:
            str: User friendly representation of the object.
        """
        return FeatureSelectionAlgorithm.to_string(self).format(
            name=self.Name, args=self._parameters_to_string(self.__de.getParameters())
        )



from niaaml.utilities import Factory
from niaaml.fitness.accuracy import Accuracy
from niaaml.fitness.cohen_kappa import CohenKappa
from niaaml.fitness.precision import Precision
from niaaml.fitness.f1 import F1

__all__ = ["FitnessFactory"]


class FitnessFactory(Factory):
    r"""Class with string mappings to fitness class.

    Attributes:
        _entities (Dict[str, Fitness]): Mapping from strings to fitness classes.

    See Also:
        * :class:`niaaml.utilities.Factory`
    """

    def _set_parameters(self, **kwargs):
        r"""Set the parameters/arguments of the factory."""
        self._entities = {
            "Accuracy": Accuracy,
            "Precision": Precision,
            "CohenKappa": CohenKappa,
            "F1": F1,
        }



from niaaml.utilities import Factory
from niaaml.classifiers.ada_boost import AdaBoost
from niaaml.classifiers.bagging import Bagging
from niaaml.classifiers.extremely_randomized_trees import ExtremelyRandomizedTrees
from niaaml.classifiers.linear_svc import LinearSVC
from niaaml.classifiers.multi_layer_perceptron import MultiLayerPerceptron
from niaaml.classifiers.random_forest import RandomForest
from niaaml.classifiers.decision_tree import DecisionTree
from niaaml.classifiers.k_neighbors import KNeighbors
from niaaml.classifiers.gaussian_process import GaussianProcess
from niaaml.classifiers.gaussian_naive_bayes import GaussianNB
from niaaml.classifiers.quadratic_driscriminant_analysis import (
    QuadraticDiscriminantAnalysis,
)

__all__ = ["ClassifierFactory"]


class ClassifierFactory(Factory):
    r"""Class with string mappings to classifiers.

    Date:
        2020

    Author:
        Luka Pečnik

    License:
        MIT

    Attributes:
        _entities (Dict[str, Classifier]): Mapping from strings to classifiers.

    See Also:
        * :class:`niaaml.utilities.Factory`
    """

    def _set_parameters(self, **kwargs):
        r"""Set the parameters/arguments of the factory."""
        self._entities = {
            "AdaBoost": AdaBoost,
            "Bagging": Bagging,
            "ExtremelyRandomizedTrees": ExtremelyRandomizedTrees,
            "LinearSVC": LinearSVC,
            "MultiLayerPerceptron": MultiLayerPerceptron,
            "RandomForest": RandomForest,
            "DecisionTree": DecisionTree,
            "KNeighbors": KNeighbors,
            "GaussianProcess": GaussianProcess,
            "GaussianNB": GaussianNB,
            "QuadraticDiscriminantAnalysis": QuadraticDiscriminantAnalysis,
        }



from niaaml.utilities import ParameterDefinition, MinMax
from niaaml.preprocessing.feature_selection.feature_selection_algorithm import (
    FeatureSelectionAlgorithm,
)
from sklearn.feature_selection import VarianceThreshold as VarThr
import numpy as np

__all__ = ["VarianceThreshold"]


class VarianceThreshold(FeatureSelectionAlgorithm):
    r"""Implementation of feature selection using variance threshold.

    Date:
        2020

    Author:
        Luka Pečnik

    License:
        MIT

    Documentation:
        https://scikit-learn.org/stable/modules/generated/sklearn.feature_selection.VarianceThreshold.html

    See Also:
        * :class:`niaaml.preprocessing.feature_selection.feature_selection_algorithm.FeatureSelectionAlgorithm`
    """
    Name = "Variance Threshold"

    def __init__(self, **kwargs):
        r"""Initialize VarianceThreshold feature selection algorithm."""
        self._params = dict(threshold=ParameterDefinition(MinMax(0, 0.1), np.float))
        self.__variance_threshold = VarThr()

    def set_parameters(self, **kwargs):
        r"""Set the parameters/arguments of the algorithm."""
        self.__variance_threshold.set_params(**kwargs)

    def select_features(self, x, y, **kwargs):
        r"""Perform the feature selection process.

        Arguments:
            x (pandas.core.frame.DataFrame): Array of original features.
            y (pandas.core.series.Series) Expected classifier results.

        Returns:
            numpy.ndarray[bool]: Mask of selected features.
        """
        self.__variance_threshold.fit(x)
        return self.__variance_threshold.get_support()

    def to_string(self):
        r"""User friendly representation of the object.

        Returns:
            str: User friendly representation of the object.
        """
        return FeatureSelectionAlgorithm.to_string(self).format(
            name=self.Name,
            args=self._parameters_to_string(self.__variance_threshold.get_params()),
        )



from niaaml.pipeline_component import PipelineComponent

__all__ = ["Classifier"]


class Classifier(PipelineComponent):
    r"""Class for implementing classifiers.

    Date:
        2020

    Author:
        Luka Pečnik

    License:
        MIT

    See Also:
        * :class:`niaaml.pipeline_component.PipelineComponent`
    """

    def fit(self, x, y, **kwargs):
        r"""Fit implemented classifier.

        Arguments:
            x (pandas.core.frame.DataFrame): n samples to classify.
            y (pandas.core.series.Series): n classes of the samples in the x array.
        """
        return

    def predict(self, x, **kwargs):
        r"""Predict class for each sample (row) in x.

        Arguments:
            x (pandas.core.frame.DataFrame): n samples to classify.

        Returns:
            pandas.core.series.Series: n predicted classes.
        """
        return



from sklearn.preprocessing import QuantileTransformer as QT
from niaaml.preprocessing.feature_transform.feature_transform_algorithm import (
    FeatureTransformAlgorithm,
)
from niaaml.utilities import ParameterDefinition

__all__ = ["QuantileTransformer"]


class QuantileTransformer(FeatureTransformAlgorithm):
    r"""Implementation of quantile transformer.

    Date:
        2020

    Author:
        Luka Pečnik

    License:
        MIT

    Documentation:
        https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.QuantileTransformer.html#sklearn.preprocessing.QuantileTransformer

    See Also:
        * :class:`niaaml.preprocessing.feature_transform.FeatureTransformAlgorithm`
    """
    Name = "Quantile Transformer"

    def __init__(self, **kwargs):
        r"""Initialize QuantileTransformer."""
        self._params = dict(
            output_distribution=ParameterDefinition(["uniform", "normal"])
        )
        self.__quantile_transformer = QT()

    def fit(self, x, **kwargs):
        r"""Fit implemented transformation algorithm.

        Arguments:
            x (pandas.core.frame.DataFrame): n samples to fit transformation algorithm.
        """
        self.__quantile_transformer.fit(x)

    def transform(self, x, **kwargs):
        r"""Transforms the given x data.

        Arguments:
            x (pandas.core.frame.DataFrame): Data to transform.

        Returns:
            pandas.core.frame.DataFrame: Transformed data.
        """

        return self.__quantile_transformer.transform(x)

    def to_string(self):
        r"""User friendly representation of the object.

        Returns:
            str: User friendly representation of the object.
        """
        return FeatureTransformAlgorithm.to_string(self).format(
            name=self.Name,
            args=self._parameters_to_string(self.__quantile_transformer.get_params()),
        )



from unittest import TestCase
from niaaml.data import BasicDataReader
import numpy


class BasicDataReaderTestCase(TestCase):
    def setUp(self):
        self.__x = numpy.random.uniform(low=0.0, high=15.0, size=(100, 6))
        self.__y = numpy.random.choice(["Class 1", "Class 2"], size=100)

    def test_x_y_works_fine(self):
        data_reader = BasicDataReader(x=self.__x, y=self.__y)
        x = data_reader.get_x()
        y = data_reader.get_y()
        self.assertEqual(x.shape, (100, 6))
        self.assertEqual(y.shape, (100,))

        self.assertTrue(numpy.all(self.__x == x))
        self.assertTrue(numpy.all(self.__y == y))

    def test_no_y_works_fine(self):
        data_reader = BasicDataReader(x=self.__x)
        x = data_reader.get_x()
        y = data_reader.get_y()
        self.assertEqual(x.shape, (100, 6))
        self.assertIsNone(y)

        self.assertTrue(numpy.all(self.__x == x))



from sklearn.model_selection import train_test_split
from niaaml.utilities import MinMax, get_bin_index, OptimizationStats
from niaaml.fitness import FitnessFactory
from niapy.benchmarks import Benchmark
from niapy.util.factory import get_algorithm
from niapy.task import StoppingTask
import pandas as pd
import numpy as np
import copy
import pickle
import os

__all__ = ["Pipeline", "_PipelineBenchmark"]


class Pipeline:
    r"""Classification pipeline defined by optional preprocessing steps and classifier.

    Date:
        2020

    Author:
        Luka Pečnik

    License:
        MIT

    Attributes:
        __feature_selection_algorithm (Optional[FeatureSelectionAlgorithm]): Feature selection algorithm implementation.
        __feature_transform_algorithm (Optional[FeatureTransformAlgorithm]): Feature transform algorithm implementation.
        __classifier (Classifier): Classifier implementation.
        __selected_features_mask (Iterable[bool]): Mask of selected features during the feature selection process.
        __best_stats (OptimizationStats): Statistics of the most successful setup of parameters.
        __categorical_features_encoders (Dict[FeatureEncoder]): Instances of FeatureEncoder for all categorical features.
        __imputers (Dict[Imputer]): Dictionary of instances of Imputer for all columns that contained missing values during optimization process.
        __logger (Logger): Logger instance.
    """

    def __init__(self, **kwargs):
        r"""Initialize task."""
        self.__feature_selection_algorithm = None
        self.__feature_transform_algorithm = None
        self.__classifier = None
        self.__selected_features_mask = None
        self.__best_stats = None
        self.__categorical_features_encoders = None
        self.__imputers = None
        self.__logger = None
        self._set_parameters(**kwargs)

    def _set_parameters(
        self,
        classifier,
        feature_selection_algorithm=None,
        feature_transform_algorithm=None,
        categorical_features_encoders=None,
        imputers=None,
        logger=None,
        **kwargs
    ):
        r"""Set the parameters/arguments of the task.

        Arguments:
            feature_selection_algorithm (Optional[FeatureSelectionAlgorithm]): Feature selection algorithm implementation.
            feature_transform_algorithm (Optional[FeatureTransformAlgorithm]): Feature transform algorithm implementation.
            classifier (Classifier): Classifier implementation.
            categorical_features_encoders (Dict[FeatureEncoders]): Actual instances of FeatureEncoder for all categorical features.
            imputers (Dict[Imputer]): Instances of Imputer for all features that contained missing values during optimization process.
            logger (Logger): Instance of the Logger class.
        """
        self.__feature_selection_algorithm = feature_selection_algorithm
        self.__feature_transform_algorithm = feature_transform_algorithm
        self.__classifier = classifier
        self.__categorical_features_encoders = categorical_features_encoders
        self.__imputers = imputers
        self.__logger = logger

    def get_feature_selection_algorithm(self):
        r"""Get deep copy of the feature selection algorithm.

        Returns:
            FeatureSelectionAlgorithm: Instance of the FeatureSelectionAlgorithm object.
        """
        return copy.deepcopy(self.__feature_selection_algorithm)

    def get_feature_transform_algorithm(self):
        r"""Get deep copy of the feature transform algorithm.

        Returns:
            FeatureTransformAlgorithm: Instance of the FeatureTransformAlgorithm object.
        """
        return copy.deepcopy(self.__feature_transform_algorithm)

    def get_classifier(self):
        r"""Get deep copy of the classifier.

        Returns:
            Classifier: Instance of the Classifier object.
        """
        return copy.deepcopy(self.__classifier)

    def get_logger(self):
        r"""Get logger.

        Returns:
            Logger: Instance of the Logger object.
        """
        return self.__logger

    def get_stats(self):
        r"""Get optimization statistics.

        Returns:
            OptimizationStats: Instance of the OptimizationStats object.
        """
        return self.__best_stats

    def set_feature_selection_algorithm(self, value):
        r"""Set feature selection algorithm."""
        self.__feature_selection_algorithm = value

    def set_feature_transform_algorithm(self, value):
        r"""Set feature transform algorithm."""
        self.__feature_transform_algorithm = value

    def set_classifier(self, value):
        r"""Set classifier."""
        self.__classifier = value

    def set_selected_features_mask(self, value):
        r"""Set selected features mask."""
        self.__selected_features_mask = value

    def set_stats(self, value):
        r"""Set stats."""
        self.__best_stats = value

    def set_categorical_features_encoders(self, value):
        r"""Set categorical features' encoders."""
        self.__categorical_features_encoders = value

    def set_imputers(self, value):
        r"""Set imputers."""
        self.__imputers = value

    def optimize(
        self,
        x,
        y,
        population_size,
        number_of_evaluations,
        optimization_algorithm,
        fitness_function,
    ):
        r"""Optimize pipeline's hyperparameters.

        Arguments:
            x (pandas.core.frame.DataFrame): n samples to classify.
            y (pandas.core.series.Series): n classes of the samples in the x array.
            population_size (uint): Number of individuals in the optimization process.
            number_of_evaluations (uint): Number of maximum evaluations.
            optimization_algorithm (str): Name of the optimization algorithm to use.
            fitness_function (str): Name of the fitness function to use.

        Returns:
            float: Best fitness value found in optimization process.
        """

        if self.__imputers is not None:
            for key in self.__imputers:
                x.loc[:, key] = self.__imputers[key].transform(x[[key]])

        if self.__categorical_features_encoders is not None:
            to_drop = []
            enc_features = pd.DataFrame()
            cols = [
                col for col in x.columns if not pd.api.types.is_numeric_dtype(x[col])
            ]
            for c in cols:
                self.__categorical_features_encoders[c].fit(x[[c]])
                tr = self.__categorical_features_encoders[c].transform(x[[c]])
                to_drop.append(c)
                enc_features = pd.concat([enc_features, tr], axis=1)
            x = x.drop(to_drop, axis=1)
            x = pd.concat([x, enc_features], axis=1)

        D = 0
        if self.__feature_selection_algorithm is not None:
            D += len(self.__feature_selection_algorithm.get_params_dict().keys())
        if self.__feature_transform_algorithm is not None:
            D += len(self.__feature_transform_algorithm.get_params_dict().keys())

        D += len(self.__classifier.get_params_dict().keys())

        algo = get_algorithm(optimization_algorithm)
        algo.NP = population_size

        task = StoppingTask(
            dimension=D,
            max_evals=number_of_evaluations,
            benchmark=_PipelineBenchmark(x, y, self, population_size, fitness_function),
        )
        best = algo.run(task)
        return best[1]

    def run(self, x):
        r"""Runs the pipeline.

        Arguments:
            x (pandas.core.frame.DataFrame): n samples to classify.

        Returns:
            pandas.core.series.Series: n predicted classes of the samples in the x array.
        """
        if self.__imputers is not None:
            for key in self.__imputers:
                x.loc[:, key] = self.__imputers[key].transform(x[[key]])

        if self.__categorical_features_encoders is not None:
            to_drop = []
            enc_features = pd.DataFrame()
            cols = [
                col for col in x.columns if not pd.api.types.is_numeric_dtype(x[col])
            ]
            for c in cols:
                tr = self.__categorical_features_encoders[c].transform(x[[c]])
                to_drop.append(c)
                enc_features = pd.concat([enc_features, tr], axis=1)
            x = x.drop(to_drop, axis=1)
            x = pd.concat([x, enc_features], axis=1)

        x = (
            x.loc[:, self.__selected_features_mask]
            if self.__selected_features_mask is not None
            else x
        )

        if self.__feature_transform_algorithm is not None:
            x = self.__feature_transform_algorithm.transform(x)

        return self.__classifier.predict(x)

    def export(self, file_name):
        r"""Exports Pipeline object to a file for later use. Extension is added if not present.

        Arguments:
            file_name (str): Output file name.
        """
        pipeline = Pipeline(
            feature_selection_algorithm=self.__feature_selection_algorithm,
            feature_transform_algorithm=self.__feature_transform_algorithm,
            classifier=self.__classifier,
            categorical_features_encoders=self.__categorical_features_encoders,
        )
        pipeline.set_selected_features_mask(self.__selected_features_mask)
        pipeline.set_stats(self.__best_stats)
        if (
            len(os.path.splitext(file_name)[1]) == 0
            or os.path.splitext(file_name)[1] != ".ppln"
        ):
            file_name = file_name + ".ppln"

        with open(file_name, "wb") as f:
            pickle.dump(pipeline, f)

    def export_text(self, file_name):
        r"""Exports Pipeline object to a user-friendly text file. Extension is added if not present.

        Arguments:
            file_name (str): Output file name.
        """
        pipeline = Pipeline(
            feature_selection_algorithm=self.__feature_selection_algorithm,
            feature_transform_algorithm=self.__feature_transform_algorithm,
            classifier=self.__classifier,
            categorical_features_encoders=self.__categorical_features_encoders,
        )
        pipeline.set_selected_features_mask(self.__selected_features_mask)
        pipeline.set_stats(self.__best_stats)
        if (
            len(os.path.splitext(file_name)[1]) == 0
            or os.path.splitext(file_name)[1] != ".txt"
        ):
            file_name = file_name + ".txt"

        with open(file_name, "w") as f:
            f.write(pipeline.to_string())

    @staticmethod
    def load(file_name):
        r"""Loads Pipeline object from a file.

        Returns:
            Pipeline: Loaded Pipeline instance.
        """
        with open(file_name, "rb") as f:
            return pickle.load(f)

    def to_string(self):
        r"""User friendly representation of the object.

        Returns:
            str: User friendly representation of the object.
        """
        classifier_string = "\t" + self.__classifier.to_string().replace("\n", "\n\t")
        feature_selection_algorithm_string = (
            "\t" + self.__feature_selection_algorithm.to_string().replace("\n", "\n\t")
            if self.__feature_selection_algorithm is not None
            else "\tNone"
        )
        feature_transform_algorithm_string = (
            "\t" + self.__feature_transform_algorithm.to_string().replace("\n", "\n\t")
            if self.__feature_transform_algorithm is not None
            else "\tNone"
        )
        stats_string = (
            "\t" + self.__best_stats.to_string().replace("\n", "\n\t")
            if self.__best_stats is not None
            else "\tStatistics is not available."
        )
        features_string = (
            "\t" + str(self.__selected_features_mask)
            if self.__selected_features_mask is not None
            else "\tFeature selection result is not available."
        )

        imputers_string = ""
        if self.__imputers is not None:
            imputers_string += "Missing features' imputers (feature's name or index: imputer's name):\n"
            for key in self.__imputers:
                imputers_string += (
                    "\t* " + str(key) + ": " + self.__imputers[key].to_string() + "\n"
                )
            imputers_string += "\n"

        encoders_string = ""
        if self.__categorical_features_encoders is not None:
            encoders_string += "Categorical features' encoders (feature's name or index: encoder's name):\n"
            for key in self.__categorical_features_encoders:
                encoders_string += (
                    "\t* "
                    + str(key)
                    + ": "
                    + self.__categorical_features_encoders[key].to_string()
                    + "\n"
                )
            encoders_string += "\n"

        return "Classifier:\n{classifier}\n\nFeature selection algorithm:\n{fsa}\n\nFeature transform algorithm:\n{fta}\n\nMask of selected features (True if selected, False if not):\n{feat}\n\n{imp}{enc}Statistics:\n{stats}".format(
            classifier=classifier_string,
            fsa=feature_selection_algorithm_string,
            fta=feature_transform_algorithm_string,
            imp=imputers_string,
            enc=encoders_string,
            feat=features_string,
            stats=stats_string,
        )

    def to_string_slim(self):
        r"""Slim user friendly representation of the object.

        Returns:
            str: Slim user friendly representation of the object.
        """

        return "classifier - {classifier}, feature selection algorithm - {fsa}, feature transform algorithm - {fta}".format(
            classifier=self.__classifier.Name,
            fsa=self.__feature_selection_algorithm.Name
            if self.__feature_selection_algorithm is not None
            else None,
            fta=self.__feature_transform_algorithm.Name
            if self.__feature_transform_algorithm is not None
            else None,
        )


class _PipelineBenchmark(Benchmark):
    r"""NiaPy Benchmark class implementation.

    Date:
        2020

    Author:
        Luka Pečnik

    License:
        MIT

    Attributes:
        __parent (Pipeline): Parent Pipeline instance.
        __x (pandas.core.frame.DataFrame): n samples to classify.
        __y (pandas.core.series.Series): n classes of the samples in the __x array.
        __population_size (uint): Number of individuals in the hiperparameter optimization process.
        __current_best_fitness (float): Current best fitness of the optimization process.
        __fitness_function (FitnessFunction): Instance of a FitnessFunction object.
        __logger (Logger): Instance of the Logger class.
        __evals (int): Number of current evaluation.
    """

    def __init__(self, x, y, parent, population_size, fitness_function):
        r"""Initialize pipeline benchmark.

        Arguments:
            parent (Pipeline): Parent instance of Pipeline.
            population_size (uint): Number of individuals in the hiperparameter optimization process.
            fitness_function (str): Name of the fitness function to use.
        """
        self.__parent = parent
        self.__x = x
        self.__y = y
        self.__population_size = population_size
        self.__current_best_fitness = float("inf")
        self.__fitness_function = FitnessFactory().get_result(fitness_function)
        self.__evals = 0
        self.__logger = self.__parent.get_logger()
        Benchmark.__init__(self, 0.0, 1.0)

    @staticmethod
    def evaluate_pipeline(
        solution_vector,
        feature_selection_algorithm,
        feature_transform_algorithm,
        classifier,
        x,
        y,
        fitness_function,
    ):
        """Evaluate pipeline setup.

        Arguments:
            solution_vector (numpy.ndarray[float]): Individual of population/ possible solution to map hyperparameters from.
            feature_selection_algorithm (Optional[FeatureSelectionAlgorithm]): Feature selection algorithm instance.
            feature_transform_algorithm (Optional[FeatureTransformAlgorithm]): Feature transform algorithm instance.
            classifier (Classifier): Classifier instance.
            x (pandas.core.frame.DataFrame): n samples to classify.
            y (pandas.core.series.Series): n classes of the samples in the x array.
            fitness_function (FitnessFunction): Fitness function instance.

        Returns:
            Tuple[float, numpy.array[bool], OptimizationStats]:
                1. Fitness.
                2. Feature selection mask.
                3. Optimization statistics.
        """
        feature_selection_algorithm_params = (
            feature_selection_algorithm.get_params_dict()
            if feature_selection_algorithm
            else dict()
        )
        feature_transform_algorithm_params = (
            feature_transform_algorithm.get_params_dict()
            if feature_transform_algorithm
            else dict()
        )
        classifier_params = classifier.get_params_dict()

        params_all = [
            (feature_selection_algorithm_params, feature_selection_algorithm),
            (feature_transform_algorithm_params, feature_transform_algorithm),
            (classifier_params, classifier),
        ]
        solution_index = 0

        for i in params_all:
            args = dict()
            for key in i[0]:
                if i[0][key] is not None:
                    if isinstance(i[0][key].value, MinMax):
                        val = (
                            solution_vector[solution_index] * i[0][key].value.max
                            + i[0][key].value.min
                        )
                        if (
                            i[0][key].param_type is np.intc
                            or i[0][key].param_type is np.int
                            or i[0][key].param_type is np.uintc
                            or i[0][key].param_type is np.uint
                        ):
                            val = i[0][key].param_type(np.floor(val))
                            if val >= i[0][key].value.max:
                                val = i[0][key].value.max - 1
                        args[key] = val
                    else:
                        args[key] = i[0][key].value[
                            get_bin_index(
                                solution_vector[solution_index], len(i[0][key].value)
                            )
                        ]
                solution_index += 1
            if i[1] is not None:
                i[1].set_parameters(**args)

        selected_features_mask = None

        x_train, x_test, y_train, y_test = train_test_split(x, y, test_size=0.2)

        if feature_selection_algorithm is None:
            selected_features_mask = np.ones(x.shape[1], dtype=bool)
        else:
            selected_features_mask = feature_selection_algorithm.select_features(
                x_train, y_train
            )

        x_train = x_train.loc[:, selected_features_mask]
        x_test = x_test.loc[:, selected_features_mask]

        if feature_transform_algorithm is not None:
            feature_transform_algorithm.fit(x_train)
            x_train = feature_transform_algorithm.transform(x_train)
            x_test = feature_transform_algorithm.transform(x_test)

        classifier.fit(x_train, y_train)
        predictions = classifier.predict(x_test)
        return (
            fitness_function.get_fitness(predictions, y_test) * -1,
            selected_features_mask,
            OptimizationStats(predictions, y_test),
        )

    def function(self):
        r"""Override Benchmark function.

        Returns:
            Callable[[int, numpy.ndarray[float]], float]: Fitness evaluation function.
        """

        def evaluate(D, sol):
            r"""Evaluate pipeline.

            Arguments:
                D (uint): Number of dimensionas.
                sol (numpy.ndarray[float]): Individual of population/ possible solution.

            Returns:
                float: Fitness.
            """
            self.__evals += 1
            if self.__logger is not None:
                self.__logger.log_progress(
                    "Evaluation {evals}".format(evals=self.__evals)
                )

            try:
                feature_selection_algorithm = (
                    self.__parent.get_feature_selection_algorithm()
                )
                feature_transform_algorithm = (
                    self.__parent.get_feature_transform_algorithm()
                )
                classifier = self.__parent.get_classifier()

                (
                    fitness,
                    selected_features_mask,
                    stats,
                ) = _PipelineBenchmark.evaluate_pipeline(
                    sol,
                    feature_selection_algorithm,
                    feature_transform_algorithm,
                    classifier,
                    self.__x,
                    self.__y,
                    self.__fitness_function,
                )

                if fitness < self.__current_best_fitness:
                    self.__current_best_fitness = fitness
                    self.__parent.set_feature_selection_algorithm(
                        feature_selection_algorithm
                    )
                    self.__parent.set_feature_transform_algorithm(
                        feature_transform_algorithm
                    )
                    self.__parent.set_classifier(classifier)
                    self.__parent.set_selected_features_mask(selected_features_mask)
                    self.__parent.set_stats(stats)

                return fitness
            except:
                # infeasible solution as it causes some kind of error
                # return infinity as we are looking for maximum accuracy in the optimization process (1 - accuracy since it is a minimization problem)
                if self.__logger is not None:
                    self.__logger.log_optimization_error(
                        "Optimization failed for: {ppln}".format(
                            ppln=self.__parent.to_string_slim()
                        )
                    )
                return float("inf")

        return evaluate



import numpy as np
from unittest import TestCase
import niaaml.fitness as f


class FitnessTestCase(TestCase):
    def setUp(self):
        self.__y = np.array(
            [
                "Class 1",
                "Class 1",
                "Class 1",
                "Class 2",
                "Class 1",
                "Class 2",
                "Class 2",
                "Class 2",
                "Class 2",
                "Class 1",
                "Class 1",
                "Class 2",
                "Class 1",
                "Class 2",
                "Class 1",
                "Class 1",
                "Class 1",
                "Class 1",
                "Class 2",
                "Class 1",
            ]
        )
        self.__predicted = np.array(
            [
                "Class 1",
                "Class 1",
                "Class 1",
                "Class 2",
                "Class 2",
                "Class 2",
                "Class 1",
                "Class 1",
                "Class 1",
                "Class 2",
                "Class 1",
                "Class 1",
                "Class 2",
                "Class 2",
                "Class 1",
                "Class 2",
                "Class 1",
                "Class 2",
                "Class 2",
                "Class 2",
            ]
        )

    def test_accuracy_works_fine(self):
        ff = f.Accuracy()
        val = ff.get_fitness(self.__predicted, self.__y)
        self.assertEqual(val, 0.5)

    def test_precision_works_fine(self):
        ff = f.Precision()
        val = ff.get_fitness(self.__predicted, self.__y)
        self.assertEqual(val, 0.5199999999999999)

    def test_cohen_kappa_works_fine(self):
        ff = f.CohenKappa()
        val = ff.get_fitness(self.__predicted, self.__y)
        self.assertEqual(val, 0.0)

    def test_f1_works_fine(self):
        ff = f.F1()
        val = ff.get_fitness(self.__predicted, self.__y)
        self.assertEqual(val, 0.505050505050505)



from niaaml.data.data_reader import DataReader
from niaaml.data.basic_data_reader import BasicDataReader
from niaaml.data.csv_data_reader import CSVDataReader

__all__ = ["DataReader", "CSVDataReader", "BasicDataReader"]



from unittest import TestCase
import niaaml.classifiers as c
from niaaml.data import CSVDataReader
import os
from sklearn.model_selection import train_test_split


class ClassifierTestCase(TestCase):
    def setUp(self):
        self.__data = CSVDataReader(
            src=os.path.dirname(os.path.abspath(__file__))
            + "/tests_files/dataset_header_classes.csv",
            has_header=True,
            contains_classes=True,
        )
        self.__x_train, self.__x_test, self.__y_train, self.__y_test = train_test_split(
            self.__data.get_x(), self.__data.get_y(), test_size=0.2
        )

    def test_adaboost_works_fine(self):
        algo = c.AdaBoost()
        algo.fit(self.__x_train, self.__y_train)
        predictions = algo.predict(self.__x_test)
        self.assertEqual(predictions.shape, self.__y_test.shape)

    def test_bagging_works_fine(self):
        algo = c.Bagging()
        algo.fit(self.__x_train, self.__y_train)
        predictions = algo.predict(self.__x_test)
        self.assertEqual(predictions.shape, self.__y_test.shape)

    def test_ert_works_fine(self):
        algo = c.ExtremelyRandomizedTrees()
        algo.fit(self.__x_train, self.__y_train)
        predictions = algo.predict(self.__x_test)
        self.assertEqual(predictions.shape, self.__y_test.shape)

    def test_lsvc_works_fine(self):
        algo = c.LinearSVC()
        algo.fit(self.__x_train, self.__y_train)
        predictions = algo.predict(self.__x_test)
        self.assertEqual(predictions.shape, self.__y_test.shape)

    def test_mlp_works_fine(self):
        algo = c.MultiLayerPerceptron()
        algo.fit(self.__x_train, self.__y_train)
        predictions = algo.predict(self.__x_test)
        self.assertEqual(predictions.shape, self.__y_test.shape)

    def test_rf_works_fine(self):
        algo = c.RandomForest()
        algo.fit(self.__x_train, self.__y_train)
        predictions = algo.predict(self.__x_test)
        self.assertEqual(predictions.shape, self.__y_test.shape)

    def test_dt_works_fine(self):
        algo = c.DecisionTree()
        algo.fit(self.__x_train, self.__y_train)
        predictions = algo.predict(self.__x_test)
        self.assertEqual(predictions.shape, self.__y_test.shape)

    def test_kn_works_fine(self):
        algo = c.KNeighbors()
        algo.fit(self.__x_train, self.__y_train)
        predictions = algo.predict(self.__x_test)
        self.assertEqual(predictions.shape, self.__y_test.shape)

    def test_gp_works_fine(self):
        algo = c.GaussianProcess()
        algo.fit(self.__x_train, self.__y_train)
        predictions = algo.predict(self.__x_test)
        self.assertEqual(predictions.shape, self.__y_test.shape)

    def test_gnb_works_fine(self):
        algo = c.GaussianNB()
        algo.fit(self.__x_train, self.__y_train)
        predictions = algo.predict(self.__x_test)
        self.assertEqual(predictions.shape, self.__y_test.shape)

    def test_qda_works_fine(self):
        algo = c.QuadraticDiscriminantAnalysis()
        algo.fit(self.__x_train, self.__y_train)
        predictions = algo.predict(self.__x_test)
        self.assertEqual(predictions.shape, self.__y_test.shape)



from niaaml.classifiers.classifier import Classifier
from sklearn.naive_bayes import GaussianNB as GNB

import warnings
from sklearn.exceptions import (
    ChangedBehaviorWarning,
    ConvergenceWarning,
    DataConversionWarning,
    DataDimensionalityWarning,
    EfficiencyWarning,
    FitFailedWarning,
    NonBLASDotWarning,
    UndefinedMetricWarning,
)

__all__ = ["GaussianNB"]


class GaussianNB(Classifier):
    r"""Implementation of gaussian Naive Bayes classifier.

    Date:
        2020

    Author:
        Luka Pečnik

    License:
        MIT

    Reference:
        Murphy, Kevin P. "Naive bayes classifiers." University of British Columbia 18 (2006): 60.

    Documentation:
        https://scikit-learn.org/stable/modules/generated/sklearn.naive_bayes.GaussianNB.html

    See Also:
        * :class:`niaaml.classifiers.Classifier`
    """
    Name = "Gaussian Naive Bayes"

    def __init__(self, **kwargs):
        r"""Initialize GaussianNB instance."""
        warnings.filterwarnings(action="ignore", category=ChangedBehaviorWarning)
        warnings.filterwarnings(action="ignore", category=ConvergenceWarning)
        warnings.filterwarnings(action="ignore", category=DataConversionWarning)
        warnings.filterwarnings(action="ignore", category=DataDimensionalityWarning)
        warnings.filterwarnings(action="ignore", category=EfficiencyWarning)
        warnings.filterwarnings(action="ignore", category=FitFailedWarning)
        warnings.filterwarnings(action="ignore", category=NonBLASDotWarning)
        warnings.filterwarnings(action="ignore", category=UndefinedMetricWarning)

        self.__gaussian_nb = GNB()
        super(GaussianNB, self).__init__()

    def set_parameters(self, **kwargs):
        r"""Set the parameters/arguments of the algorithm."""
        self.__gaussian_nb.set_params(**kwargs)

    def fit(self, x, y, **kwargs):
        r"""Fit GaussianNB.

        Arguments:
            x (pandas.core.frame.DataFrame): n samples to classify.
            y (pandas.core.series.Series): n classes of the samples in the x array.

        Returns:
            None
        """
        self.__gaussian_nb.fit(x, y)

    def predict(self, x, **kwargs):
        r"""Predict class for each sample (row) in x.

        Arguments:
            x (pandas.core.frame.DataFrame): n samples to classify.

        Returns:
            pandas.core.series.Series: n predicted classes.
        """
        return self.__gaussian_nb.predict(x)

    def to_string(self):
        r"""User friendly representation of the object.

        Returns:
            str: User friendly representation of the object.
        """
        return Classifier.to_string(self).format(
            name=self.Name,
            args=self._parameters_to_string(self.__gaussian_nb.get_params()),
        )



from sklearn.metrics import accuracy_score
from niaaml.fitness.fitness_function import FitnessFunction

__all__ = ["Accuracy"]


class Accuracy(FitnessFunction):
    r"""Class representing the accuracy as a fitness function.

    Date:
        2020

    Author:
        Luka Pečnik

    License:
        MIT

    Documentation:
        https://scikit-learn.org/stable/modules/generated/sklearn.metrics.accuracy_score.html

    See Also:
        * :class:`niaaml.fitness.FitnessFunction`
    """
    Name = "Accuracy"

    def get_fitness(self, predicted, expected):
        r"""Return fitness value. The larger return value should represent a better fitness for the framework to work properly.

        Arguments:
            predicted (pandas.core.series.Series): Predicted values.
            expected (pandas.core.series.Series): Expected values.

        Returns:
            float: Calculated fitness value.
        """
        return accuracy_score(expected, predicted)



from niapy.benchmarks import Benchmark
from sklearn.model_selection import train_test_split
from sklearn.linear_model import LogisticRegression

__all__ = ["_FeatureSelectionThresholdBenchmark"]


class _FeatureSelectionThresholdBenchmark(Benchmark):
    r"""NiaPy Benchmark class implementation.

    Attributes:
        __best_fitness (float): Current best fitness of the optimization process.
        __best_solution (numpy.ndarray[float]): Current best solution of the optimization process.
    """

    def __init__(self, X, y):
        r"""Initialize feature selection benchmark.

        Arguments:
            X (pandas.core.frame.DataFrame): Features.
            y (pandas.core.series.Series) Expected classifier results.
        """
        self.__best_fitness = float("inf")
        self.__best_solution = None
        Benchmark.__init__(self, 0.0, 1.0)
        self.train_X, self.test_X, self.train_y, self.test_y = train_test_split(
            X, y, test_size=0.2
        )

    def get_best_solution(self):
        r"""Get best solution found.

        Returns:
            numpy.ndarray[float]: Best solution found.
        """
        return self.__best_solution

    def function(self):
        r"""Override Benchmark function.

        Returns:
            Callable[[int, numpy.ndarray[float]], float]: Fitness evaluation function.
        """

        def evaluate(D, sol):
            r"""Evaluate features.

            Arguments:
                D (uint): Number of dimensions.
                sol (numpy.ndarray[float]): Individual of population/ possible solution.

            Returns:
                float: Fitness.
            """
            selected = []  # array for selected features
            self.Threshold = sol[D - 1]  # current threshold

            # select features
            for i in range(sol.shape[0] - 1):
                if sol[i] < self.Threshold:
                    pass
                else:
                    selected.append(i)

            # in the case if threshold is too low (no features selected)
            if len(selected) == 0:
                return 1

            lr = LogisticRegression(solver="lbfgs", max_iter=10000).fit(
                self.train_X.iloc[:, selected], self.train_y
            )
            accuracy = lr.score(self.test_X.iloc[:, selected], self.test_y)
            fitness = 1.0 - accuracy

            if fitness < self.__best_fitness:
                self.__best_fitness = fitness
                self.__best_solution = sol
            return fitness

        return evaluate



__all__ = ["PipelineComponent"]


class PipelineComponent:
    r"""Class for implementing pipeline components.

    Date:
        2020

    Author:
        Luka Pečnik

    License:
        MIT

    Attributes:
        Name (str): Name of the pipeline component.
        _params (Dict[str, ParameterDefinition]): Dictionary of components's parameters with possible values. Possible parameter values are given as an instance of the ParameterDefinition class.

    See Also:
        * :class:`niaaml.utilities.ParameterDefinition`
    """
    Name = None

    def __init__(self, **kwargs):
        r"""Initialize pipeline component.

        Notes:
            _params variable should not be static as in some cases it is instance specific. See * :class:`niaaml.preprocessing.feature_selection.select_k_best.SelectKBest` for example.
        """
        self._params = dict()
        self.set_parameters(**kwargs)

    def set_parameters(self, **kwargs):
        r"""Set the parameters/arguments of the pipeline component."""
        return

    def get_params_dict(self):
        r"""Return parameters definition dictionary."""
        return self._params

    def to_string(self):
        r"""User friendly representation of the object.

        Returns:
            str: User friendly representation of the object.
        """
        return "Name: {name}\nArguments:\n{args}"

    def _parameters_to_string(self, dictionary):
        r"""User friendly representation of component's parameters.

        Arguments:
            dictionary (dict): Dictionary of parameters.

        Returns:
            str: User friendly representation of component's parameters.
        """
        args_string = ""
        for key in dictionary:
            args_string += "\t" + key + " = " + str(dictionary[key]) + "\n"
        if len(args_string) == 0:
            args_string = "None"
        return args_string



from sklearn.preprocessing import Normalizer as Nrm
from niaaml.preprocessing.feature_transform.feature_transform_algorithm import (
    FeatureTransformAlgorithm,
)
from niaaml.utilities import ParameterDefinition

__all__ = ["Normalizer"]


class Normalizer(FeatureTransformAlgorithm):
    r"""Implementation of feature normalization algorithm.

    Date:
        2020

    Author:
        Luka Pečnik

    License:
        MIT

    Documentation:
        https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.Normalizer

    See Also:
        * :class:`niaaml.preprocessing.feature_transform.FeatureTransformAlgorithm`
    """
    Name = "Normalizer"

    def __init__(self, **kwargs):
        r"""Initialize Normalizer."""
        self._params = dict(norm=ParameterDefinition(["l1", "l2", "max"]))
        self.__params = None
        self.__normalizer = Nrm()

    def set_parameters(self, **kwargs):
        r"""Set the parameters/arguments of the algorithm."""
        self.__params = kwargs
        self.__params["axis"] = 0

    def fit(self, x, **kwargs):
        r"""Fit implemented transformation algorithm.

        Arguments:
            x (pandas.core.frame.DataFrame): n samples to fit transformation algorithm.
        """
        self.__normalizer.fit(x)

    def transform(self, x, **kwargs):
        r"""Transforms the given x data.

        Arguments:
            x (pandas.core.frame.DataFrame): Data to transform.

        Returns:
            pandas.core.frame.DataFrame: Transformed data.
        """
        return self.__normalizer.transform(x)

    def to_string(self):
        r"""User friendly representation of the object.

        Returns:
            str: User friendly representation of the object.
        """
        return FeatureTransformAlgorithm.to_string(self).format(
            name=self.Name,
            args=self._parameters_to_string(self.__normalizer.get_params()),
        )



from unittest import TestCase
from niaaml import ParameterDefinition, MinMax, OptimizationStats, get_bin_index
import numpy as np
import tempfile


class UtilitiesTestCase(TestCase):
    def test_get_bin_index_works_fine(self):
        self.assertEqual(get_bin_index(0.0, 4), 0)
        self.assertEqual(get_bin_index(0.24, 4), 0)
        self.assertEqual(get_bin_index(0.25, 4), 1)
        self.assertEqual(get_bin_index(0.49, 4), 1)
        self.assertEqual(get_bin_index(0.5, 4), 2)
        self.assertEqual(get_bin_index(0.74, 4), 2)
        self.assertEqual(get_bin_index(0.75, 4), 3)
        self.assertEqual(get_bin_index(1.0, 4), 3)


class ParameterDefinitionTestCase(TestCase):
    def test_works_fine(self):
        parameter_definition = ParameterDefinition(MinMax(0.0, 5.9), float)

        self.assertIsInstance(parameter_definition.value, MinMax)
        self.assertEqual(parameter_definition.param_type, float)


class OptimizationStatsTestCase(TestCase):
    def setUp(self):
        y = np.array(
            [
                "Class 1",
                "Class 1",
                "Class 1",
                "Class 2",
                "Class 1",
                "Class 2",
                "Class 2",
                "Class 2",
                "Class 2",
                "Class 1",
                "Class 1",
                "Class 2",
                "Class 1",
                "Class 2",
                "Class 1",
                "Class 1",
                "Class 1",
                "Class 1",
                "Class 2",
                "Class 1",
            ]
        )
        predicted = np.array(
            [
                "Class 1",
                "Class 1",
                "Class 1",
                "Class 2",
                "Class 2",
                "Class 2",
                "Class 1",
                "Class 1",
                "Class 1",
                "Class 2",
                "Class 1",
                "Class 1",
                "Class 2",
                "Class 2",
                "Class 1",
                "Class 2",
                "Class 1",
                "Class 2",
                "Class 2",
                "Class 2",
            ]
        )

        self.__stats = OptimizationStats(predicted, y)

    def test_works_fine(self):
        self.assertEqual(self.__stats._accuracy, 0.5)
        self.assertEqual(self.__stats._precision, 0.5199999999999999)
        self.assertEqual(self.__stats._cohen_kappa, 0.0)
        self.assertEqual(self.__stats._f1_score, 0.505050505050505)


class MinMaxTestCase(TestCase):
    def test_works_fine(self):
        minmax = MinMax(0.0, 5.9)

        self.assertEqual(minmax.min, 0.0)
        self.assertEqual(minmax.max, 5.9)



from niaaml.pipeline import Pipeline, _PipelineBenchmark
from niaaml.classifiers import ClassifierFactory
from niaaml.preprocessing.feature_selection import FeatureSelectionAlgorithmFactory
from niaaml.preprocessing.feature_transform import FeatureTransformAlgorithmFactory
from niapy.task import StoppingTask
from niapy.benchmarks import Benchmark
from niapy.util.factory import get_algorithm
from niaaml.utilities import get_bin_index
from niaaml.preprocessing.encoding.utility import encode_categorical_features
from niaaml.preprocessing.imputation.utility import impute_features
from niaaml.fitness import FitnessFactory
from niaaml.logger import Logger

__all__ = [
    "PipelineOptimizer",
    "_PipelineOptimizerBenchmark",
    "_PipelineOptimizerBenchmarkV1",
    "_PipelineOptimizerBenchmarkV2",
]


class PipelineOptimizer:
    r"""Optimization task that finds the best classification pipeline according to the given input.

    Date:
        2020

    Author:
        Luka Pečnik

    License:
        MIT

    Attributes:
        __data (DataReader): Instance of any DataReader implementation.
        __feature_selection_algorithms (Optional[Iterable[str]]): Array of names of possible feature selection algorithms.
        __feature_transform_algorithms (Optional[Iterable[str]]): Array of names of possible feature transform algorithms.
        __classifiers (Iterable[Classifier]): Array of names of possible classifiers.
        __categorical_features_encoder (str): Name of the encoder used for categorical features.
        __categorical_features_encoders (Dict[FeatureEncoder]): Actual instances of FeatureEncoder for all categorical features.
        __imputer (str): Name of the imputer used for features that contain missing values.
        __imputers (Dict[Imputer]): Actual instances of Imputer for all features that contain missing values.
        __logger (Logger): Logger instance.
    """

    def __init__(self, **kwargs):
        r"""Initialize task."""
        self.__data = None
        self.__feature_selection_algorithms = None
        self.__feature_transform_algorithms = None
        self.__classifiers = None
        self.__categorical_features_encoder = None
        self.__categorical_features_encoders = None
        self.__imputer = None
        self.__imputers = None
        self.__logger = None

        self._set_parameters(**kwargs)

    def _set_parameters(
        self,
        data,
        classifiers,
        feature_selection_algorithms=None,
        feature_transform_algorithms=None,
        categorical_features_encoder=None,
        imputer=None,
        log=True,
        log_verbose=False,
        log_output_file=None,
        **kwargs
    ):
        r"""Set the parameters/arguments of the task.

        Arguments:
            data (DataReader): Instance of any DataReader implementation.
            feature_selection_algorithms (Optional[Iterable[str]]): Array of names of possible feature selection algorithms.
            feature_transform_algorithms (Optional[Iterable[str]]): Array of names of possible feature transform algorithms.
            classifiers (Iterable[Classifier]): Array of names of possible classifiers.
            categorical_features_encoder (Optional[str]): Name of the encoder used for categorical features.
            imputer (Optional[str]): Name of the imputer used for features that contain missing values.
            log (Optional(bool)): Log optimization progress.
            log_verbose (Optional(bool)): Log optimization progress, current best pipeline and warnings.
            log_output_file (Optional(str)): Path to the logging output file. Defaults to standard output.
        """
        self.__data = data

        self.__feature_transform_algorithms = feature_transform_algorithms
        if self.__feature_transform_algorithms is not None:
            try:
                self.__feature_transform_algorithms.index(None)
            except:
                self.__feature_transform_algorithms.insert(0, None)

        self.__classifiers = classifiers
        self.__feature_selection_algorithms = feature_selection_algorithms
        self.__categorical_features_encoder = categorical_features_encoder
        self.__imputer = imputer

        if log is True:
            self.__logger = Logger(verbose=log_verbose, output_file=log_output_file)

    def get_data(self):
        r"""Get data.

        Returns:
            DataReader: Instance of DataReader object.
        """
        return self.__data

    def get_feature_selection_algorithms(self):
        r"""Get feature selection algorithms.

        Returns:
            Iterable[str]: Feature selection algorithm names or None.
        """
        return self.__feature_selection_algorithms

    def get_feature_transform_algorithms(self):
        r"""Get feature transform algorithms.

        Returns:
            Iterable[str]: Feature transform algorithm names or None.
        """
        return self.__feature_transform_algorithms

    def get_classifiers(self):
        r"""Get classifiers.

        Returns:
            Iterable[str]: Classifier names.
        """
        return self.__classifiers

    def get_logger(self):
        r"""Get logger.

        Returns:
            Logger: Logger instance.
        """
        return self.__logger

    def run(
        self,
        fitness_name,
        pipeline_population_size,
        inner_population_size,
        number_of_pipeline_evaluations,
        number_of_inner_evaluations,
        optimization_algorithm,
        inner_optimization_algorithm=None,
    ):
        r"""Run classification pipeline optimization process.

        Arguments:
            fitness_name (str): Name of the fitness class to use as a function.
            pipeline_population_size (uint): Number of pipeline individuals in the optimization process.
            inner_population_size (uint): Number of individuals in the hiperparameter optimization process.
            number_of_pipeline_evaluations (uint): Number of maximum evaluations.
            number_of_inner_evaluations (uint): Number of maximum inner evaluations.
            optimization_algorithm (str): Name of the optimization algorithm to use.
            inner_optimization_algorithm (Optional[str]): Name of the inner optimization algorithm to use. Defaults to the optimization_algorithm argument.

        Returns:
            Pipeline: Best pipeline found in the optimization process.
        """

        algo = get_algorithm(optimization_algorithm)
        algo.NP = pipeline_population_size

        features = self.__data.get_x()

        if self.__imputer is not None:
            features, self.__imputers = impute_features(features, self.__imputer)

        if self.__categorical_features_encoder is not None:
            (
                features,
                self.__categorical_features_encoders,
            ) = encode_categorical_features(
                features, self.__categorical_features_encoder
            )

        self.__data.set_x(features)

        benchmark = _PipelineOptimizerBenchmarkV2(
            self,
            fitness_name,
            inner_population_size,
            number_of_inner_evaluations,
            inner_optimization_algorithm
            if inner_optimization_algorithm is not None
            else optimization_algorithm,
        )
        task = StoppingTask(
            dimension=3, max_evals=number_of_pipeline_evaluations, benchmark=benchmark
        )
        algo.run(task)

        pipeline = benchmark.get_pipeline()
        if pipeline is not None:
            pipeline.set_categorical_features_encoders(
                self.__categorical_features_encoders
            )
            pipeline.set_imputers(self.__imputers)

        return pipeline

    def run_v1(
        self,
        fitness_name,
        population_size,
        number_of_evaluations,
        optimization_algorithm,
    ):
        r"""Run classification pipeline optimization process according to the original NiaAML paper.

        Reference:
            Fister, Iztok, Milan Zorman, and Dušan Fister. "Continuous Optimizers for Automatic Design and Evaluation of Classification Pipelines." Frontier Applications of Nature Inspired Computation. Springer, Singapore, 2020. 281-301.

        Arguments:
            fitness_name (str): Name of the fitness class to use as a function.
            population_size (uint): Number of individuals in the optimization process.
            number_of_evaluations (uint): Number of maximum evaluations.
            optimization_algorithm (str): Name of the optimization algorithm to use.

        Returns:
            Pipeline: Best pipeline found in the optimization process.
        """

        algo = get_algorithm(optimization_algorithm)
        algo.NP = population_size

        features = self.__data.get_x()

        if self.__imputer is not None:
            features, self.__imputers = impute_features(features, self.__imputer)

        if self.__categorical_features_encoder is not None:
            (
                features,
                self.__categorical_features_encoders,
            ) = encode_categorical_features(
                features, self.__categorical_features_encoder
            )

        self.__data.set_x(features)

        D = 3
        factories = [
            (self.__feature_selection_algorithms, FeatureSelectionAlgorithmFactory()),
            (self.__feature_transform_algorithms, FeatureTransformAlgorithmFactory()),
            (self.__classifiers, ClassifierFactory()),
        ]

        for f in factories:
            m = 0
            if f[0] is not None:
                for e in f[0]:
                    if e is not None:
                        el = f[1].get_result(e)
                        params = el.get_params_dict()
                        if params is not None and len(params) > m:
                            m = len(params)
            D += m

        benchmark = _PipelineOptimizerBenchmarkV1(self, fitness_name)
        task = StoppingTask(
            dimension=D, max_evals=number_of_evaluations, benchmark=benchmark
        )
        algo.run(task)

        pipeline = benchmark.get_pipeline()
        if pipeline is not None:
            pipeline.set_categorical_features_encoders(
                self.__categorical_features_encoders
            )
            pipeline.set_imputers(self.__imputers)

        return pipeline


class _PipelineOptimizerBenchmark(Benchmark):
    r"""NiaPy Benchmark class base implementation.

    Date:
        2020

    Author:
        Luka Pečnik

    Attributes:
        _parent (PipelineOptimizer): Parent instance of PipelineOptimizer.
        _current_best_fitness (float): Current best fitness of the optimization process.
        _current_best_pipeline (Pipeline): Current best pipeline of the optimization process.
        _fitness_name (str): Name of the fitness class to use as a function.
        _logger (Logger): Instance of the Logger class.
        _evals (int): Number of current evaluation.

        _classifier_factory (ClassifierFactory): Factory for classifiers.
        _feature_transform_algorithm_factory (FeatureTransformAlgorithmFactory): Factory for feature transform algorithms.
        _feature_selection_algorithm_factory (FeatureSelectionAlgorithmFactory): Factory for feature selection algorithms.
    """
    _classifier_factory = ClassifierFactory()
    _feature_transform_algorithm_factory = FeatureTransformAlgorithmFactory()
    _feature_selection_algorithm_factory = FeatureSelectionAlgorithmFactory()

    def __init__(self, parent, fitness_name):
        r"""Initialize pipeline optimizer benchmark.

        Arguments:
            parent (PipelineOptimizer): Parent instance of PipelineOptimizer.
            fitness_name (str): Name of the fitness class to use as a function.
        """
        self._parent = parent
        self._current_best_fitness = float("inf")
        self._current_best_pipeline = None
        self._fitness_name = fitness_name
        self._evals = 0
        self._logger = self._parent.get_logger()
        Benchmark.__init__(self, 0.0, 1.0)

    def _float_to_instance(self, value, collection, factory):
        r"""Get instance of object from collection using factory.

        Arguments:
            value (float): Value to map.
            collection (Iterable[str]): Array of names of possible feature selection algorithms.
            factory (Factory): Implementation of the Factory class.

        Returns:
            PipelineComponent: New PipelineComponent instance.
        """
        bin_index = get_bin_index(value, len(collection))

        name = collection[bin_index]
        return factory.get_result(name) if name is not None else None

    def get_pipeline(self):
        r"""Get best pipeline found.

        Returns:
            Pipeline: Best pipeline found.
        """
        return self._current_best_pipeline


class _PipelineOptimizerBenchmarkV2(_PipelineOptimizerBenchmark):
    r"""NiaPy Benchmark class implementation.

    Date:
        2020

    Author:
        Luka Pečnik

    Attributes:
        __inner_population_size (uint): Number of individuals in the hiperparameter optimization process.
        __number_of_inner_evaluations (uint): Number of maximum inner evaluations.
        __optimization_algorithm (str): Name of the optimization algorithm to use.

    See also:
        * :class:`niaaml.pipeline_optimizer._PipelineOptimizerBenchmark`
    """

    def __init__(
        self,
        parent,
        fitness_name,
        inner_population_size,
        number_of_inner_evaluations,
        inner_optimization_algorithm,
    ):
        r"""Initialize pipeline optimizer benchmark.

        Arguments:
            parent (PipelineOptimizer): Parent instance of PipelineOptimizer.
            fitness_name (str): Name of the fitness class to use as a function.
            inner_population_size (uint): Number of individuals in the hiperparameter optimization process.
            number_of_inner_evaluations (uint): Number of maximum inner evaluations.
            inner_optimization_algorithm (str): Name of the optimization algorithm to use.
        """
        super(_PipelineOptimizerBenchmarkV2, self).__init__(parent, fitness_name)
        self.__inner_population_size = inner_population_size
        self.__number_of_inner_evaluations = number_of_inner_evaluations
        self.__optimization_algorithm = inner_optimization_algorithm

    def function(self):
        r"""Override Benchmark function.

        Returns:
            Callable[[int, numpy.ndarray[float]], float]: Fitness evaluation function.
        """

        def evaluate(D, sol):
            r"""Evaluate pipeline.

            Arguments:
                D (uint): Number of dimensions.
                sol (numpy.ndarray[float]): Individual of population/ possible solution.

            Returns:
                float: Fitness.
            """
            self._evals += 1

            data = self._parent.get_data()
            fs_algo = (
                self._float_to_instance(
                    sol[0],
                    self._parent.get_feature_selection_algorithms(),
                    self._feature_selection_algorithm_factory,
                )
                if self._parent.get_feature_selection_algorithms() is not None
                and len(self._parent.get_feature_selection_algorithms()) > 0
                else None
            )
            ft_algo = (
                self._float_to_instance(
                    sol[1],
                    self._parent.get_feature_transform_algorithms(),
                    self._feature_transform_algorithm_factory,
                )
                if self._parent.get_feature_transform_algorithms() is not None
                and len(self._parent.get_feature_transform_algorithms()) > 0
                else None
            )
            clsf = self._float_to_instance(
                sol[2], self._parent.get_classifiers(), self._classifier_factory
            )

            pipeline = Pipeline(
                feature_selection_algorithm=fs_algo,
                feature_transform_algorithm=ft_algo,
                classifier=clsf,
                logger=self._logger,
            )

            if self._logger is not None:
                self._logger.log_progress(
                    "Currently optimizing {evals}: {ppln}".format(
                        ppln=pipeline.to_string_slim(), evals=self._evals
                    )
                )

            fitness = pipeline.optimize(
                data.get_x(),
                data.get_y(),
                self.__inner_population_size,
                self.__number_of_inner_evaluations,
                self.__optimization_algorithm,
                self._fitness_name,
            )
            if fitness < self._current_best_fitness:
                self._current_best_fitness = fitness
                self._current_best_pipeline = pipeline

                if self._logger is not None:
                    self._logger.log_pipeline(
                        "New current best pipeline with fitness {fit}: {ppln}".format(
                            fit=-fitness, ppln=pipeline.to_string_slim()
                        )
                    )

            return fitness

        return evaluate


class _PipelineOptimizerBenchmarkV1(_PipelineOptimizerBenchmark):
    r"""NiaPy Benchmark class implementation.

    Date:
        2020

    Author:
        Luka Pečnik

    See also:
        * :class:`niaaml.pipeline_optimizer._PipelineOptimizerBenchmark`
    """

    def __init__(self, parent, fitness_name):
        r"""Initialize pipeline optimizer benchmark.

        Arguments:
            parent (PipelineOptimizer): Parent instance of PipelineOptimizer.
            fitness_name (str): Name of the fitness class to use as a function.
        """
        super(_PipelineOptimizerBenchmarkV1, self).__init__(parent, fitness_name)
        self.__fitness_function = FitnessFactory().get_result(self._fitness_name)

    def function(self):
        r"""Override Benchmark function.

        Returns:
            Callable[[int, numpy.ndarray[float]], float]: Fitness evaluation function.
        """

        def evaluate(D, sol):
            r"""Evaluate pipeline.

            Arguments:
                D (uint): Number of dimensions.
                sol (numpy.ndarray[float]): Individual of population/ possible solution.

            Returns:
                float: Fitness.
            """
            self._evals += 1
            if self._logger is not None:
                self._logger.log_progress(
                    "Evaluation {evals}".format(evals=self._evals)
                )

            try:
                data = self._parent.get_data()
                fs_algo = (
                    self._float_to_instance(
                        sol[0],
                        self._parent.get_feature_selection_algorithms(),
                        self._feature_selection_algorithm_factory,
                    )
                    if self._parent.get_feature_selection_algorithms() is not None
                    and len(self._parent.get_feature_selection_algorithms()) > 0
                    else None
                )
                ft_algo = (
                    self._float_to_instance(
                        sol[1],
                        self._parent.get_feature_transform_algorithms(),
                        self._feature_transform_algorithm_factory,
                    )
                    if self._parent.get_feature_transform_algorithms() is not None
                    and len(self._parent.get_feature_transform_algorithms()) > 0
                    else None
                )
                clsf = self._float_to_instance(
                    sol[2], self._parent.get_classifiers(), self._classifier_factory
                )

                pipeline = Pipeline(
                    feature_selection_algorithm=fs_algo,
                    feature_transform_algorithm=ft_algo,
                    classifier=clsf,
                    logger=self._logger,
                )

                if self._logger is not None:
                    self._logger.log_progress(
                        "Currently optimizing: {ppln}".format(
                            ppln=pipeline.to_string_slim()
                        )
                    )

                (
                    fitness,
                    selected_features_mask,
                    stats,
                ) = _PipelineBenchmark.evaluate_pipeline(
                    sol[3:],
                    fs_algo,
                    ft_algo,
                    clsf,
                    data.get_x(),
                    data.get_y(),
                    self.__fitness_function,
                )

                if fitness < self._current_best_fitness:
                    self._current_best_fitness = fitness
                    pipeline.set_feature_selection_algorithm(fs_algo)
                    pipeline.set_feature_transform_algorithm(ft_algo)
                    pipeline.set_classifier(clsf)
                    pipeline.set_stats(stats)
                    pipeline.set_selected_features_mask(selected_features_mask)
                    self._current_best_pipeline = pipeline

                    if self._logger is not None:
                        self._logger.log_pipeline(
                            "New current best pipeline with fitness {fit}: {ppln}".format(
                                fit=-fitness, ppln=pipeline.to_string_slim()
                            )
                        )

                return fitness
            except:
                # infeasible solution as it causes some kind of error
                # return infinity as we are looking for maximum accuracy in the optimization process (1 - accuracy since it is a minimization problem)
                if self._logger is not None:
                    self._logger.log_optimization_error(
                        "Optimization failed for: {ppln}".format(
                            ppln=pipeline.to_string_slim()
                        )
                    )
                return float("inf")

        return evaluate



from unittest import TestCase
import niaaml.preprocessing.feature_selection as fs
from niaaml.data import CSVDataReader
import os


class FeatureSelectionTestCase(TestCase):
    def setUp(self):
        self.__data = CSVDataReader(
            src=os.path.dirname(os.path.abspath(__file__))
            + "/tests_files/dataset_header_classes.csv",
            has_header=True,
            contains_classes=True,
        )

    def test_PSO_works_fine(self):
        algo = fs.ParticleSwarmOptimization()
        selected_features_mask = algo.select_features(
            self.__data.get_x(), self.__data.get_y()
        )
        self.assertEqual(self.__data.get_x().shape[1], len(selected_features_mask))

    def test_select_k_best_works_fine(self):
        algo = fs.SelectKBest()
        selected_features_mask = algo.select_features(
            self.__data.get_x(), self.__data.get_y()
        )
        self.assertEqual(self.__data.get_x().shape[1], len(selected_features_mask))

    def test_select_percentile_works_fine(self):
        algo = fs.SelectPercentile()
        selected_features_mask = algo.select_features(
            self.__data.get_x(), self.__data.get_y()
        )
        self.assertEqual(self.__data.get_x().shape[1], len(selected_features_mask))

    def test_bat_algorithm_works_fine(self):
        algo = fs.BatAlgorithm()
        selected_features_mask = algo.select_features(
            self.__data.get_x(), self.__data.get_y()
        )
        self.assertEqual(self.__data.get_x().shape[1], len(selected_features_mask))

    def test_de_works_fine(self):
        algo = fs.DifferentialEvolution()
        selected_features_mask = algo.select_features(
            self.__data.get_x(), self.__data.get_y()
        )
        self.assertEqual(self.__data.get_x().shape[1], len(selected_features_mask))

    def test_gwo_works_fine(self):
        algo = fs.GreyWolfOptimizer()
        selected_features_mask = algo.select_features(
            self.__data.get_x(), self.__data.get_y()
        )
        self.assertEqual(self.__data.get_x().shape[1], len(selected_features_mask))

    def test_jdefsth_works_fine(self):
        algo = fs.jDEFSTH()
        selected_features_mask = algo.select_features(
            self.__data.get_x(), self.__data.get_y()
        )
        self.assertEqual(self.__data.get_x().shape[1], len(selected_features_mask))

    def test_vt_works_fine(self):
        algo = fs.VarianceThreshold()
        selected_features_mask = algo.select_features(
            self.__data.get_x(), self.__data.get_y()
        )
        self.assertEqual(self.__data.get_x().shape[1], len(selected_features_mask))



from sklearn.metrics import accuracy_score, precision_score, cohen_kappa_score, f1_score
import numpy as np

__all__ = [
    "MinMax",
    "ParameterDefinition",
    "OptimizationStats",
    "Factory",
    "get_bin_index",
]


def get_bin_index(value, number_of_bins):
    """Gets index of value's bin. Value must be between 0.0 and 1.0.

    Arguments:
        value (float): Value to put into bin.
        number_of_bins (uint): Number of bins on the interval [0.0, 1.0].

    Returns:
        uint: Calculated index.
    """
    bin_index = np.int(np.floor(value / (1.0 / number_of_bins)))
    if bin_index >= number_of_bins:
        bin_index -= 1
    return bin_index


class MinMax:
    r"""Class for ParameterDefinition's value property.

    Date:
        2020

    Author:
        Luka Pečnik

    License:
        MIT

    Attributes:
        min (float): Minimum number (inclusive).
        max (float): Maximum number (exclusive).

    See Also:
        * :class:`niaaml.utilities.ParameterDefinition`
    """

    def __init__(self, min, max):
        r"""Initialize instance.

        Arguments:
            min (float): Minimum number (inclusive).
            max (float): Maximum number (exclusive).
        """
        self.min = min
        self.max = max


class ParameterDefinition:
    r"""Class for PipelineComponent parameters definition.

    Date:
        2020

    Author:
        Luka Pečnik

    License:
        MIT

    Attributes:
        value (any): Array of possible parameter values or instance of MinMax class.
        param_type (numpy.dtype): Selection output data type.

    See Also:
        * :class:`niaaml.pipeline_component.PipelineComponent`
        * :class:`niaaml.utilities.MinMax`
    """

    def __init__(self, value, param_type=None):
        r"""Initialize instance.

        Arguments:
            value (Any): Array of any type or instance of MinMax class.
            param_type (Optional[type]): Type of possible outcome according to an instance of MinMax class. Not used if the argument value is array.
        """
        self.value = value
        self.param_type = param_type


class Factory:
    r"""Base class with string mappings to entities.

    Date:
        2020

    Author:
        Luka Pečnik

    License:
        MIT

    Attributes:
        _entities (Dict[str, any]): Dictionary to map from strings to an instance of anything.
    """
    _entities = None

    def __init__(self, **kwargs):
        r"""Initialize the factory."""
        self._set_parameters(**kwargs)

    def _set_parameters(self, **kwargs):
        r"""Set the parameters/arguments of the factory."""
        return

    def get_result(self, name):
        r"""Get the resulting entity.

        Arguments:
            name (str): String that represents the entity.

        Returns:
            any: Entity according to the given name.
        """

        if name in self._entities:
            return self._entities[name]()
        else:
            raise TypeError("Passed entity is not defined! --> %s" % name)

    def get_name_to_classname_mapping(self):
        r"""Get dictionary of user-friendly name to class name mapping.

        Returns:
            dict: Dictionary of user-friendly name to class name mapping.
        """
        d = {}
        for k in self._entities:
            d[self._entities[k].Name] = k
        return d


class OptimizationStats:
    r"""Class that holds pipeline optimization result's statistics. Includes accuracy, precision, Cohen's kappa and F1-score.

    Date:
        2020

    Author:
        Luka Pečnik

    License:
        MIT

    Attributes:
        _accuracy (float): Calculated accuracy.
        _precision (float): Calculated precision.
        _cohen_kappa (float): Calculated Cohen's kappa.
        _f1_score (float): Calculated F1-score.
    """

    def __init__(self, predicted, expected, **kwargs):
        r"""Initialize the factory.

        Arguments:
            predicted (Iterable[any]): Array of predicted classes.
            expected (Iterable[any]): Array of expected classes.
        """
        self._accuracy = accuracy_score(expected, predicted)
        self._precision = precision_score(expected, predicted, average="weighted")
        self._cohen_kappa = cohen_kappa_score(expected, predicted)
        self._f1_score = f1_score(expected, predicted, average="weighted")

    def to_string(self):
        r"""User friendly representation of the object.

        Returns:
            str: User friendly representation of the object.
        """
        return "Accuracy: {acc},\nPrecision: {prc},\nCohen's kappa: {ck},\nF1-score: {f1}".format(
            acc=self._accuracy,
            prc=self._precision,
            ck=self._cohen_kappa,
            f1=self._f1_score,
        )



from unittest import TestCase
from niaaml.data import CSVDataReader
import os


class CSVDataReaderTestCase(TestCase):
    def test_header_classes_works_fine(self):
        data_reader = CSVDataReader(
            src=os.path.dirname(os.path.abspath(__file__))
            + "/tests_files/dataset_header_classes.csv",
            has_header=True,
            contains_classes=True,
        )
        x = data_reader.get_x()
        y = data_reader.get_y()
        self.assertEqual(x.shape, (100, 6))
        self.assertEqual(y.shape, (100,))

    def test_no_header_classes_works_fine(self):
        data_reader = CSVDataReader(
            src=os.path.dirname(os.path.abspath(__file__))
            + "/tests_files/dataset_no_header_classes.csv",
            has_header=False,
            contains_classes=True,
        )
        x = data_reader.get_x()
        y = data_reader.get_y()
        self.assertEqual(x.shape, (100, 6))
        self.assertEqual(y.shape, (100,))

    def test_no_header_no_classes_works_fine(self):
        data_reader = CSVDataReader(
            src=os.path.dirname(os.path.abspath(__file__))
            + "/tests_files/dataset_no_header_no_classes.csv",
            has_header=False,
            contains_classes=False,
        )
        x = data_reader.get_x()
        y = data_reader.get_y()
        self.assertEqual(x.shape, (100, 6))
        self.assertIsNone(y)

    def test_header_no_classes_works_fine(self):
        data_reader = CSVDataReader(
            src=os.path.dirname(os.path.abspath(__file__))
            + "/tests_files/dataset_header_no_classes.csv",
            has_header=True,
            contains_classes=False,
        )
        x = data_reader.get_x()
        y = data_reader.get_y()
        self.assertEqual(x.shape, (100, 6))
        self.assertIsNone(y)



from niaaml.preprocessing.imputation.imputer import Imputer
from niaaml.preprocessing.imputation.simple_imputer import SimpleImputer
from niaaml.preprocessing.imputation.utility import ImputerFactory
from niaaml.preprocessing.imputation.utility import impute_features

__all__ = ["Imputer", "SimpleImputer", "ImputerFactory", "impute_features"]
